import datetime
import json
from pathlib import Path
from typing import Any

from loguru import logger

from climate_ref_core.datasets import FacetFilter, SourceDatasetType
from climate_ref_core.diagnostics import (
    CommandLineDiagnostic,
    DataRequirement,
    ExecutionDefinition,
    ExecutionResult,
)
from climate_ref_core.pycmec.metric import remove_dimensions
from climate_ref_pmp.pmp_driver import build_glob_pattern, build_pmp_command, process_json_result

# =================================================================
# PMP diagnostics support functions for the annual cycle diagnostic
# =================================================================


def make_data_requirement(variable_id: str, obs_source: str) -> tuple[DataRequirement, DataRequirement]:
    """
    Create a data requirement for the annual cycle diagnostic.

    Parameters
    ----------
    variable_id : str
        The variable ID to filter the data requirement.
    obs_source : str
        The observation source ID to filter the data requirement.

    Returns
    -------
    DataRequirement
        A DataRequirement object containing the necessary filters and groupings.
    """
    return (
        DataRequirement(
            source_type=SourceDatasetType.PMPClimatology,
            filters=(FacetFilter(facets={"source_id": (obs_source,), "variable_id": (variable_id,)}),),
            group_by=("variable_id", "source_id"),
        ),
        DataRequirement(
            source_type=SourceDatasetType.CMIP6,
            filters=(
                FacetFilter(
                    facets={
                        "frequency": "mon",
                        "experiment_id": ("amip", "historical", "hist-GHG"),
                        "variable_id": (variable_id,),
                    }
                ),
            ),
            group_by=("variable_id", "source_id", "experiment_id", "member_id", "grid_label"),
        ),
    )


def _transform_results(data: dict[str, Any]) -> dict[str, Any]:
    """
    Transform the executions dictionary to match the expected structure.

    Parameters
    ----------
    data : dict
        The original execution dictionary.

    Returns
    -------
    dict
        The transformed executions dictionary.
    """
    # Remove the model, reference, rip dimensions
    # These are later replaced with a REF-specific naming convention
    data = remove_dimensions(data, ["model", "reference", "rip"])

    # TODO: replace this with the ability to capture series
    # Remove the "CalendarMonths" key from the nested structure
    for region, region_values in data["RESULTS"].items():
        for stat, stat_values in region_values.items():
            if "CalendarMonths" in stat_values:
                stat_values.pop("CalendarMonths")

    # Remove the "CalendarMonths" key from the nested structure in "DIMENSIONS"
    data["DIMENSIONS"]["season"].pop("CalendarMonths")

    return data


def transform_results_files(results_files: list[Any]) -> list[Any]:
    """
    Transform the results files to match the expected structure.

    Parameters
    ----------
    results_files : list
        List of result files to transform.

    Returns
    -------
    list
        List of transformed result files.

    """
    if len(results_files) == 0:
        logger.warning("No results files provided for transformation.")
        return []

    transformed_results_files = []

    for results_file in results_files:
        # Rewrite the CMEC JSON file for compatibility
        with open(results_file) as f:
            results = json.load(f)
            results_transformed = _transform_results(results)

        # Get the stem (filename without extension)
        stem = results_file.stem

        # Create the new filename
        results_file_transformed = results_file.with_name(f"{stem}_transformed.json")

        with open(results_file_transformed, "w") as f:
            # Write the transformed executions back to the file
            json.dump(results_transformed, f, indent=4)
            logger.debug(f"Transformed executions written to {results_file_transformed}")

        transformed_results_files.append(results_file_transformed)

    return transformed_results_files


def _update_top_level_keys(combined_results: dict[str, Any], data: dict[str, Any], levels: list[str]) -> None:
    if "DIMENSIONS" not in data:
        data["DIMENSIONS"] = {}

    top_level_keys = list(data.keys())
    top_level_keys.remove("RESULTS")

    json_structure = data.get("DIMENSIONS", {}).get("json_structure", {})
    json_structure = ["level", *json_structure]

    for key in top_level_keys:
        combined_results[key] = data[key]
        if key == "Variable":
            combined_results[key]["level"] = levels
        elif key == "DIMENSIONS":
            combined_results[key]["json_structure"] = json_structure
            if "level" not in combined_results[key]:
                combined_results[key]["level"] = {}
                for level in levels:
                    combined_results[key]["level"][level] = {}


def combine_results_files(results_files: list[Any], output_directory: str | Path) -> Path:
    """
    Combine multiple results files into a single file.

    Parameters
    ----------
    results_files : list
        List of result files to combine.
    output_directory : str or Path
        Directory where the combined file will be saved.

    Returns
    -------
    Path, list[str]
        The path to the combined results file and a list of levels found in the results files.
    """
    combined_results: dict[str, dict[str, dict[str, dict[str, dict[str, Any]]]]] = {}
    combined_results["RESULTS"] = {}
    levels = []

    # Ensure output_directory is a Path object
    if isinstance(output_directory, str):
        output_directory = Path(output_directory)

    last_data = None
    for file in results_files:
        with open(file) as f:
            data = json.load(f)
            last_data = data
            level_key = str(int(data["Variable"]["level"]))
            levels.append(level_key)
            logger.debug(f"Processing file: {file}, level_key: {level_key}")
            # Insert the results into the combined_results dictionary
            if level_key not in combined_results["RESULTS"]:
                combined_results["RESULTS"][level_key] = data.get("RESULTS", {})

    if last_data is not None:
        _update_top_level_keys(combined_results, last_data, levels)

    # Ensure the output directory exists
    output_directory.mkdir(parents=True, exist_ok=True)

    # Create the combined file path
    combined_file_path = output_directory / "combined_results.json"

    with open(combined_file_path, "w") as f:
        json.dump(combined_results, f, indent=4)

    # return combined_file_path, levels
    return combined_file_path


# ===================================================
# PMP diagnostics main class: annual cycle diagnostic
# ===================================================


class AnnualCycle(CommandLineDiagnostic):
    """
    Calculate the annual cycle for a dataset
    """

    name = "Annual Cycle"
    slug = "annual-cycle"
    facets = (
        "source_id",
        "member_id",
        "experiment_id",
        "variable_id",
        "reference_source_id",
        "region",
        "statistic",
        "season",
    )

    data_requirements = (
        # ERA-5 as reference dataset, spatial 2-D variables
        make_data_requirement("ts", "ERA-5"),
        make_data_requirement("uas", "ERA-5"),
        make_data_requirement("vas", "ERA-5"),
        make_data_requirement("psl", "ERA-5"),
        # ERA-5 as reference dataset, spatial 3-D variables
        make_data_requirement("ta", "ERA-5"),
        make_data_requirement("ua", "ERA-5"),
        make_data_requirement("va", "ERA-5"),
        make_data_requirement("zg", "ERA-5"),
        # Other reference datasets, spatial 2-D variables
        make_data_requirement("pr", "GPCP-Monthly-3-2"),
        make_data_requirement("rlds", "CERES-EBAF-4-2"),
        make_data_requirement("rlus", "CERES-EBAF-4-2"),
        make_data_requirement("rlut", "CERES-EBAF-4-2"),
        make_data_requirement("rsds", "CERES-EBAF-4-2"),
        make_data_requirement("rsdt", "CERES-EBAF-4-2"),
        make_data_requirement("rsus", "CERES-EBAF-4-2"),
        make_data_requirement("rsut", "CERES-EBAF-4-2"),
    )

    def __init__(self) -> None:
        self.parameter_file_1 = "pmp_param_annualcycle_1-clims.py"
        self.parameter_file_2 = "pmp_param_annualcycle_2-metrics.py"

    def build_cmds(self, definition: ExecutionDefinition) -> list[list[str]]:
        """
        Build the command to run the diagnostic

        Parameters
        ----------
        definition
            Definition of the diagnostic execution

        Returns
        -------
            Command arguments to execute in the PMP environment
        """
        input_datasets = definition.datasets[SourceDatasetType.CMIP6]
        reference_datasets = definition.datasets[SourceDatasetType.PMPClimatology]

        source_id = input_datasets["source_id"].unique()[0]
        experiment_id = input_datasets["experiment_id"].unique()[0]
        member_id = input_datasets["member_id"].unique()[0]
        variable_id = input_datasets["variable_id"].unique()[0]

        model_files_raw = input_datasets.path.to_list()
        if len(model_files_raw) == 1:
            model_files = model_files_raw[0]  # If only one file, use it directly
        elif len(model_files_raw) > 1:
            model_files = build_glob_pattern(model_files_raw)  # If multiple files, build a glob pattern
        else:
            raise ValueError("No model files found")

        logger.debug("build_cmd start")

        logger.debug(f"input_datasets: {input_datasets}")
        logger.debug(f"input_datasets.keys(): {input_datasets.keys()}")

        reference_dataset_name = reference_datasets["source_id"].unique()[0]
        reference_dataset_path = reference_datasets.datasets.iloc[0]["path"]

        logger.debug(f"reference_dataset.datasets: {reference_datasets.datasets}")
        logger.debug(f"reference_dataset_name: {reference_dataset_name}")
        logger.debug(f"reference_dataset_path: {reference_dataset_path}")

        output_directory_path = str(definition.output_directory)

        cmds = []

        # ----------------------------------------------
        # PART 1: Build the command to get climatologies
        # ----------------------------------------------
        # Model
        data_name = f"{source_id}_{experiment_id}_{member_id}"
        data_path = model_files
        params = {
            "vars": variable_id,
            "infile": data_path,
            "outfile": f"{output_directory_path}/{variable_id}_{data_name}_clims.nc",
        }

        cmds.append(
            build_pmp_command(
                driver_file="pcmdi_compute_climatologies.py",
                parameter_file=self.parameter_file_1,
                **params,
            )
        )

        # --------------------------------------------------
        # PART 2: Build the command to calculate diagnostics
        # --------------------------------------------------
        # Reference
        obs_dict = {
            variable_id: {
                reference_dataset_name: {
                    "template": reference_dataset_path,
                },
                "default": reference_dataset_name,
            }
        }

        # Generate a JSON file based on the obs_dict
        with open(f"{output_directory_path}/obs_dict.json", "w") as f:
            json.dump(obs_dict, f)

        date = datetime.datetime.now().strftime("%Y%m%d")

        if variable_id in ["ua", "va", "ta"]:
            levels = ["200", "850"]
        elif variable_id in ["zg"]:
            levels = ["500"]
        else:
            levels = None

        variables = []
        if levels is not None:
            for level in levels:
                variable_id_with_level = f"{variable_id}-{level}"
                variables.append(variable_id_with_level)
        else:
            variables = [variable_id]

        logger.debug(f"variables: {variables}")
        logger.debug(f"levels: {levels}")

        # Build the command for each level
        params = {
            "vars": variables,
            "custom_observations": f"{output_directory_path}/obs_dict.json",
            "test_data_path": output_directory_path,
            "test_data_set": source_id,
            "realization": member_id,
            "filename_template": f"%(variable)_{data_name}_clims.198101-200512.AC.v{date}.nc",
            "metrics_output_path": output_directory_path,
            "cmec": "",
        }

        cmds.append(
            build_pmp_command(
                driver_file="mean_climate_driver.py",
                parameter_file=self.parameter_file_2,
                **params,
            )
        )

        logger.debug("build_cmd end")
        logger.debug(f"cmds: {cmds}")

        return cmds

    def build_execution_result(self, definition: ExecutionDefinition) -> ExecutionResult:
        """
        Build a diagnostic result from the output of the PMP driver

        Parameters
        ----------
        definition
            Definition of the diagnostic execution

        Returns
        -------
            Result of the diagnostic execution
        """
        input_datasets = definition.datasets[SourceDatasetType.CMIP6]
        variable_id = input_datasets["variable_id"].unique()[0]

        if variable_id in ["ua", "va", "ta"]:
            variable_dir_pattern = f"{variable_id}-???"
        else:
            variable_dir_pattern = variable_id

        results_directory = definition.output_directory
        png_directory = results_directory / variable_dir_pattern
        data_directory = results_directory / variable_dir_pattern

        logger.debug(f"results_directory: {results_directory}")
        logger.debug(f"png_directory: {png_directory}")
        logger.debug(f"data_directory: {data_directory}")

        # Find the CMEC JSON file(s)
        results_files = transform_results_files(list(results_directory.glob("*_cmec.json")))

        if len(results_files) == 1:
            # If only one file, use it directly
            results_file = results_files[0]
            logger.debug(f"results_file: {results_file}")
        elif len(results_files) > 1:
            logger.info(f"More than one cmec file found: {results_files}")
            results_file = combine_results_files(results_files, definition.output_directory)
        else:
            logger.error("Unexpected case: no cmec file found")
            return ExecutionResult.build_from_failure(definition)

        # Find the other outputs: PNG and NetCDF files
        png_files = list(png_directory.glob("*.png"))
        data_files = list(data_directory.glob("*.nc"))

        # Prepare the output bundles
        cmec_output_bundle, cmec_metric_bundle = process_json_result(results_file, png_files, data_files)

        # Add missing dimensions to the output
        input_selectors = input_datasets.selector_dict()
        reference_selectors = definition.datasets[SourceDatasetType.PMPClimatology].selector_dict()
        cmec_metric_bundle = cmec_metric_bundle.prepend_dimensions(
            {
                "source_id": input_selectors["source_id"],
                "member_id": input_selectors["member_id"],
                "experiment_id": input_selectors["experiment_id"],
                "variable_id": input_selectors["variable_id"],
                "reference_source_id": reference_selectors["source_id"],
            }
        )

        return ExecutionResult.build_from_output_bundle(
            definition,
            cmec_output_bundle=cmec_output_bundle,
            cmec_metric_bundle=cmec_metric_bundle,
        )

    def execute(self, definition: ExecutionDefinition) -> None:
        """
        Run the diagnostic on the given configuration.

        Parameters
        ----------
        definition : ExecutionDefinition
            The configuration to run the diagnostic on.

        Returns
        -------
        :
            The result of running the diagnostic.
        """
        cmds = self.build_cmds(definition)

        runs = [self.provider.run(cmd) for cmd in cmds]
        logger.debug(f"runs: {runs}")



"""
ENSO diagnostic driver

This script runs inside the PMP conda environment due to the use of xcdat.
"""

import argparse
import copy
import json
import os
from collections import defaultdict

import xcdat as xc

from pcmdi_metrics import resources  # isort:skip
from pcmdi_metrics.enso.lib import metrics_to_json  # isort:skip
from pcmdi_metrics.io import StringConstructor  # isort:skip
from pcmdi_metrics.utils import create_land_sea_mask

from EnsoMetrics.EnsoCollectionsLib import defCollection  # isort:skip
from EnsoMetrics.EnsoComputeMetricsLib import ComputeCollection  # isort:skip
from EnsoPlots.EnsoMetricPlot import main_plotter  # isort:skip


def main():
    """
    Run the ENSO metrics collection and plotting.

    This script is designed to be run from the command line.
    It takes two command line arguments:
    1. input_json_path: Path to the JSON file containing the datasets.
    2. output_directory: Directory where the output files will be saved.
    """
    print("### PMP ENSO: Compute the metric collection ###\n")

    args = parse_arguments()
    dict_datasets, mod, run, pattern = prepare_datasets(args)
    dict_metric, dict_dive = compute_metrics(args, dict_datasets, mod, run, pattern)
    save_metrics_to_json(args, dict_datasets, dict_metric, dict_dive, pattern)
    plot_results(args, pattern, mod, run)


def parse_arguments():
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(description="A script that takes two inputs and processes them.")
    parser.add_argument("--metrics_collection", type=str, help="metrics collection")
    parser.add_argument("--experiment_id", type=str, help="experiment id")
    parser.add_argument("--input_json_path", type=str, help="JSON file path")
    parser.add_argument("--output_directory", type=str, help="output directory")
    return parser.parse_args()


def prepare_datasets(args):
    """Prepare datasets and update them with land-sea masks."""
    os.makedirs(args.output_directory, exist_ok=True)
    with open(args.input_json_path) as f:
        dict_datasets = json.load(f)
    mod_run = next(iter(dict_datasets["model"].keys()))
    mod, run = mod_run.split("_")
    pattern = f"{args.metrics_collection}_{mod}_{args.experiment_id}_{run}"
    dict_datasets = update_dict_datasets(dict_datasets, os.path.join(args.output_directory, "ref_landmask"))
    # Write a JSON file for dict_datasets
    json_file = os.path.join(args.output_directory, f"input_{pattern}_processed.json")
    with open(json_file, "w") as f:
        json.dump(dict_datasets, f, indent=4)
    print(f"JSON file created: {json_file}")
    return dict_datasets, mod, run, pattern


def compute_metrics(args, dict_datasets, mod, run, pattern):
    """Compute the metric collection."""
    dict_metric = defaultdict(dict)
    dict_dive = defaultdict(dict)
    metrics, dive_results = ComputeCollection(
        args.metrics_collection,
        dict_datasets,
        f"{mod}_{run}",
        netcdf=True,
        netcdf_name=os.path.join(args.output_directory, pattern),
        obs_interpreter=True,
        debug=True,
    )
    dict_metric[mod][run] = metrics
    dict_dive[mod][run] = dive_results
    return dict_metric, dict_dive


def save_metrics_to_json(args, dict_datasets, dict_metric, dict_dive, pattern):
    """Save metrics to a JSON file."""
    egg_pth = resources.resource_path()
    dict_obs = dict_datasets["observations"]
    # pattern = f"{args.metrics_collection}_{mod}_{args.experiment_id}_{run}"
    mod = pattern.split("_")[-3]
    run = pattern.split("_")[-1]
    metrics_to_json(
        args.metrics_collection,
        dict_obs,
        dict_metric,
        dict_dive,
        egg_pth,
        StringConstructor(args.output_directory),
        pattern,
        mod=mod,
        run=run,
    )
    # Write an additional JSON file for the results for CMEC standard
    json_file = os.path.join(args.output_directory, f"{pattern}.json")
    write_CMEC_json(json_file)


def write_CMEC_json(json_file):
    """
    Write the CMEC JSON file.

    Parameters
    ----------
    json_file : str
        Path to the input JSON file.
    """
    # Load the existing JSON file
    with open(json_file) as f:
        dict_data = json.load(f)

    # -----------------------------------------
    # Prepare components for the CMEC structure
    # -----------------------------------------
    metrics_dict = {}
    ref_datasets = []

    mod = next(iter(dict_data["RESULTS"]["model"].keys()))
    run = next(iter(dict_data["RESULTS"]["model"][mod].keys()))

    metrics = list(dict_data["RESULTS"]["model"][mod][run]["value"].keys())
    for metric in metrics:
        metrics_dict[metric] = {}
        ref_datasets.extend(list(dict_data["RESULTS"]["model"][mod][run]["value"][metric]["metric"].keys()))

    ref_datasets = list(set(ref_datasets))  # Remove duplicates

    ref_datasets_dict = {ref: {} for ref in ref_datasets}

    dimensions_dict = {
        "json_structure": ["model", "realization", "metric", "reference_datasets"],
        "model": {mod: {}},
        "realization": {run: {}},
        "metric": metrics_dict,
        "reference_datasets": ref_datasets_dict,
    }

    results_dict = {}
    results_dict[mod] = {}
    results_dict[mod][run] = {}

    for metric in metrics:
        results_dict[mod][run][metric] = {}
        ref_datasets = list(dict_data["RESULTS"]["model"][mod][run]["value"][metric]["metric"].keys())
        for ref_dataset in ref_datasets:
            value = dict_data["RESULTS"]["model"][mod][run]["value"][metric]["metric"][ref_dataset]["value"]
            results_dict[mod][run][metric][ref_dataset] = value

    # Check if the dimensions are in the JSON file
    if "provenance" in dict_data:
        provenance_dict = dict_data["provenance"]
    else:
        provenance_dict = {}

    # Check if the reference datasets are in the JSON file
    if "obs" in dict_data["RESULTS"]:
        ref_dict = dict_data["RESULTS"]["obs"]
    else:
        ref_dict = {}

    # -----------------------------------------------
    # Create a new dictionary with the CMEC structure
    # -----------------------------------------------
    cmec_dict = {
        "RESULTS": results_dict,
        "DIMENSIONS": dimensions_dict,
        "REF": ref_dict,
        "provenance": provenance_dict,
    }

    # ---------------------------------------
    # Write the new dictionary to a JSON file
    # ---------------------------------------
    json_cmec_file = json_file.replace(".json", "_cmec.json")

    with open(json_cmec_file, "w") as f:
        json.dump(cmec_dict, f, indent=4)


def plot_results(args, pattern, mod, run):
    """Plot the results."""
    mod_run = f"{mod}_{run}"
    with open(os.path.join(args.output_directory, f"{pattern}.json")) as ff:
        data_json = json.load(ff)["RESULTS"]["model"][mod][run]
    plot_enso(
        args.metrics_collection,
        mod_run,
        args.experiment_id,
        args.output_directory,
        data_json,
    )


def plot_enso(mc_name, mod_run, exp, path_work_dir, data_json):
    """
    Plot the ENSO metrics collection.

    Parameters
    ----------
    mc_name : str
        Name of the metrics collection.
    mod_run : str
        Model and run name, separated by an underscore.
        e.g., "ACCESS-CM2_r1i1p1f1".
    exp : str
        Experiment name.
    path_work_dir : str
        Path of directory that contains the input NetCDF files and used to save the output PNG files.
    data_json : dict
        Data loaded from the JSON file.
    """
    metrics = sorted(defCollection(mc_name)["metrics_list"].keys(), key=lambda v: v.upper())
    print(f"metrics: {metrics}")

    mod = mod_run.split("_")[0]
    run = mod_run.split("_")[1]

    pattern = "_".join([mc_name, mod, exp, run])

    for met in metrics:
        print(f"met: {met}")
        # get NetCDF file name
        filename_nc = os.path.join(path_work_dir, pattern + "_" + met + ".nc")
        print(f"filename_nc: {filename_nc}")
        if os.path.exists(filename_nc):
            # get diagnostic values for the given model and observations
            if mc_name == "ENSO_tel" and "Map" in met:
                dict_dia = data_json["value"][met + "Corr"]["diagnostic"]
                diagnostic_values = dict((key1, None) for key1 in dict_dia.keys())
                diagnostic_units = ""
            else:
                dict_dia = data_json["value"][met]["diagnostic"]
                diagnostic_values = dict((key1, dict_dia[key1]["value"]) for key1 in dict_dia.keys())
                diagnostic_units = data_json["metadata"]["metrics"][met]["diagnostic"]["units"]
            # get metric values computed with the given model and observations
            if mc_name == "ENSO_tel" and "Map" in met:
                list1, list2 = (
                    [met + "Corr", met + "Rmse"],
                    [
                        "diagnostic",
                        "metric",
                    ],
                )
                dict_met = data_json["value"]
                metric_values = dict(
                    (
                        key1,
                        {mod: [dict_met[su][ty][key1]["value"] for su, ty in zip(list1, list2)]},
                    )
                    for key1 in dict_met[list1[0]]["metric"].keys()
                )
                metric_units = [data_json["metadata"]["metrics"][su]["metric"]["units"] for su in list1]
            else:
                dict_met = data_json["value"][met]["metric"]
                metric_values = dict((key1, {mod: dict_met[key1]["value"]}) for key1 in dict_met.keys())
                metric_units = data_json["metadata"]["metrics"][met]["metric"]["units"]
            # figure name
            figure_name = "_".join([mc_name, mod, exp, run, met])
            print(f"figure_name: {figure_name}")

            main_plotter(
                mc_name,
                met,
                mod,
                exp,
                filename_nc,
                diagnostic_values,
                diagnostic_units,
                metric_values,
                metric_units,
                member=run,
                path_png=path_work_dir,
                name_png=figure_name,
            )

            print("figure plotting done")

        else:
            print(f"file not found: {filename_nc}")


def update_dict_datasets(dict_datasets: dict, output_dir: str = ".") -> dict:
    """
    Update the dictDatasets to include the land-sea mask and remap observation names.

    Parameters
    ----------
    dict_datasets : dict
        Dictionary containing datasets information.
    output_dir : str
        Directory where the land-sea mask will be saved.
        Default is the current directory.

    Returns
    -------
    dict
        Updated dictionary with land-sea mask and remapped observation names.

    Raises
    ------
    FileNotFoundError
        If the input file path is not valid.
    NotImplementedError
        If multiple paths are found for a dataset or if the path is not a string.
    """
    dict_datasets2 = copy.deepcopy(dict_datasets)
    data_types = dict_datasets.keys()  # ["model", "observations"]

    # Select only model and observations datasets
    data_types = [data_type for data_type in data_types if data_type in ["model", "observations"]]

    for data_type in data_types:
        datasets = dict_datasets[data_type].keys()
        for dataset in datasets:
            variables = dict_datasets[data_type][dataset].keys()
            for variable in variables:
                path = dict_datasets[data_type][dataset][variable]["path + filename"]

                # If path is a list and has one element, take it as a string,
                # otherwise raise notImplementedError
                if isinstance(path, list) and len(path) == 1:
                    path = copy.deepcopy(path[0])
                    dict_datasets2[data_type][dataset][variable]["path + filename"] = path
                elif isinstance(path, list) and len(path) > 1:
                    raise NotImplementedError(
                        f"Multiple paths found for {data_type} {dataset} {variable}: {path}"
                    )
                elif not isinstance(path, str):
                    raise NotImplementedError(
                        f"Path is not a string for {data_type} {dataset} {variable}: {path}"
                    )
                else:
                    dict_datasets2[data_type][dataset][variable]["path + filename"] = path

                # Check if the file exists
                if not os.path.exists(path):
                    raise FileNotFoundError(f"File not found: {path}")

                # Generate the landmask path regardless data_type is observation or model.
                if (
                    "path + filename_area" not in dict_datasets[data_type][dataset]
                    or "path + filename_landmask" not in dict_datasets[data_type][dataset]
                ):
                    # Generate it per variable as different variables may be on different grids.
                    path_landmask = generate_landmask_path(path, variable, output_dir=output_dir)

                    dict_datasets2[data_type][dataset][variable]["areaname"] = "areacella"
                    dict_datasets2[data_type][dataset][variable]["landmaskname"] = "sftlf"
                    dict_datasets2[data_type][dataset][variable]["path + filename_area"] = path_landmask
                    dict_datasets2[data_type][dataset][variable]["path + filename_landmask"] = path_landmask

                # Map variable names to ENSO package recognized names
                var_name_mapping = {"ts": "sst", "tauu": "taux"}
                var_name_key = var_name_mapping.get(variable, variable)

                # Update the variable name
                dict_datasets2[data_type][dataset][var_name_key] = dict_datasets2[data_type][dataset].pop(
                    variable
                )

            if data_type == "observations":
                # Mapping of old observation names to new ones recognized by the ENSO package
                observation_name_mapping = {
                    "GPCP-2-3": "GPCPv2.3",
                    "ERA-INT": "ERA-Interim",
                    "ERA-5": "ERA5",
                    "AVISO-1-0": "AVISO",
                    "TropFlux-1-0": "Tropflux",
                    "HadISST-1-1": "HadISST",
                }
                # Get the new name if it exists in the mapping, otherwise keep the original name
                dataset_name_key = observation_name_mapping.get(dataset, dataset)
                # Update the dictDatasets with the new name
                dict_datasets2[data_type][dataset_name_key] = dict_datasets2[data_type].pop(dataset)

    return dict_datasets2


def generate_landmask_path(file_path, var_name, output_dir=".", output_filename=None):
    """
    Generate the landmask path based on the given file path.

    Parameters
    ----------
    file_path : str
        Path to the input NetCDF file.
    var_name : str
        Variable name to be used for creating the land-sea mask.
    output_dir : str
        Directory where the land-sea mask will be saved.
        Default is the current directory.
    output_filename : str
        Name of the output land-sea mask file.
        If not provided, it will be generated based on the input file name.
        Default is None.

    Returns
    -------
    str
        Path to the generated land-sea mask file.

    Raises
    ------
    FileNotFoundError
        If the input file path is not valid.
    ValueError
        If the variable name is not valid.
    """
    # If file_path is a list, take the first element
    if isinstance(file_path, list):
        file_path = file_path[0]

    # Check if the file path is valid
    if not os.path.isfile(file_path):
        raise FileNotFoundError(f"File not found: {file_path}")

    # Check if the variable name is valid
    if not isinstance(var_name, str):
        raise ValueError(f"Invalid variable name: {var_name}")

    # Open the dataset using xcdat and create the land-sea mask
    ds = xc.open_dataset(file_path)
    mask = create_land_sea_mask(ds[var_name])

    # Name mask variable as 'sftlf'
    mask.name = "sftlf"

    # Check if the output directory exists, create it if not
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # Save the land-sea mask to a new NetCDF file
    if output_filename:
        landmask_filename = output_filename
    else:
        # Generate the filename based on the input file name
        landmask_filename = os.path.basename(file_path).replace(".nc", "_landmask.nc")

    landmask_path = os.path.join(output_dir, landmask_filename)
    mask.to_netcdf(landmask_path)

    return os.path.abspath(landmask_path)


if __name__ == "__main__":
    main()



"""
Rapid evaluating CMIP data
"""

from __future__ import annotations

import importlib.metadata
import os
from typing import TYPE_CHECKING

from loguru import logger

from climate_ref_core.dataset_registry import DATASET_URL, dataset_registry_manager
from climate_ref_core.providers import CondaDiagnosticProvider
from climate_ref_pmp.diagnostics import ENSO, AnnualCycle, ExtratropicalModesOfVariability

if TYPE_CHECKING:
    from climate_ref.config import Config

__version__ = importlib.metadata.version("climate-ref-pmp")


# Create the PMP diagnostics provider
# PMP uses a conda environment to run the diagnostics
class PMPDiagnosticProvider(CondaDiagnosticProvider):
    """
    Provider for PMP diagnostics.
    """

    def configure(self, config: Config) -> None:
        """Configure the provider."""
        super().configure(config)
        self.env_vars["PCMDI_CONDA_EXE"] = str(self.get_conda_exe())
        # This is a workaround for a fatal error in internal_Finalize of MPICH
        # when running in a conda environment on MacOS.
        # It is not clear if this is a bug in MPICH or a problem with the conda environment.
        if "FI_PROVIDER" not in os.environ:  # pragma: no branch
            logger.debug("Setting env variable 'FI_PROVIDER=tcp'")
            self.env_vars["FI_PROVIDER"] = "tcp"


provider = PMPDiagnosticProvider("PMP", __version__)


# Annual cycle diagnostics and metrics
provider.register(AnnualCycle())

# ENSO diagnostics and metrics
# provider.register(ENSO("ENSO_perf"))  # Assigned to ESMValTool
provider.register(ENSO("ENSO_tel"))
provider.register(ENSO("ENSO_proc"))

# Extratropical modes of variability diagnostics and metrics
provider.register(ExtratropicalModesOfVariability("PDO"))
provider.register(ExtratropicalModesOfVariability("NPGO"))
provider.register(ExtratropicalModesOfVariability("NAO"))
provider.register(ExtratropicalModesOfVariability("NAM"))
provider.register(ExtratropicalModesOfVariability("PNA"))
provider.register(ExtratropicalModesOfVariability("NPO"))
provider.register(ExtratropicalModesOfVariability("SAM"))


dataset_registry_manager.register(
    "pmp-climatology",
    base_url=DATASET_URL,
    package="climate_ref_pmp.dataset_registry",
    resource="pmp_climatology.txt",
)



"""PMP diagnostics."""

from climate_ref_pmp.diagnostics.annual_cycle import AnnualCycle
from climate_ref_pmp.diagnostics.enso import ENSO
from climate_ref_pmp.diagnostics.variability_modes import ExtratropicalModesOfVariability

__all__ = [
    "ENSO",
    "AnnualCycle",
    "ExtratropicalModesOfVariability",
]



import datetime

import pandas as pd
import pytest
from climate_ref_pmp import AnnualCycle
from climate_ref_pmp import provider as pmp_provider
from climate_ref_pmp.diagnostics.annual_cycle import _transform_results
from climate_ref_pmp.pmp_driver import _get_resource

from climate_ref.solver import solve_executions
from climate_ref_core.datasets import DatasetCollection, SourceDatasetType
from climate_ref_core.diagnostics import ExecutionDefinition


def test_expected_executions():
    diagnostic = AnnualCycle()
    data_catalog = {
        SourceDatasetType.CMIP6: pd.DataFrame(
            [
                ["ts", "ACCESS-ESM1-5", "historical", "r1i1p1f1", "mon", "gn"],
                ["ts", "ACCESS-ESM1-5", "ssp119", "r1i1p1f1", "mon", "gn"],
                ["ts", "ACCESS-ESM1-5", "historical", "r2i1p1f1", "mon", "gn"],
                ["pr", "ACCESS-ESM1-5", "historical", "r1i1p1f1", "mon", "gn"],
            ],
            columns=("variable_id", "source_id", "experiment_id", "member_id", "frequency", "grid_label"),
        ),
        SourceDatasetType.PMPClimatology: pd.DataFrame(
            [["ERA-5", "ts"], ["ERA-5", "pr"], ["GPCP-Monthly-3-2", "pr"]],
            columns=["source_id", "variable_id"],
        ),
    }
    executions = list(solve_executions(data_catalog, diagnostic, provider=pmp_provider))
    assert len(executions) == 3

    # ts
    assert executions[0].datasets[SourceDatasetType.CMIP6].selector == (
        ("experiment_id", "historical"),
        ("grid_label", "gn"),
        ("member_id", "r1i1p1f1"),
        ("source_id", "ACCESS-ESM1-5"),
        ("variable_id", "ts"),
    )
    assert executions[0].datasets[SourceDatasetType.PMPClimatology].selector == (
        ("source_id", "ERA-5"),
        ("variable_id", "ts"),
    )

    # ts with different member_id
    assert executions[1].datasets[SourceDatasetType.CMIP6].selector == (
        ("experiment_id", "historical"),
        ("grid_label", "gn"),
        ("member_id", "r2i1p1f1"),
        ("source_id", "ACCESS-ESM1-5"),
        ("variable_id", "ts"),
    )
    assert executions[0].datasets[SourceDatasetType.PMPClimatology].selector == (
        ("source_id", "ERA-5"),
        ("variable_id", "ts"),
    )

    # pr
    assert executions[2].datasets[SourceDatasetType.CMIP6].selector == (
        ("experiment_id", "historical"),
        ("grid_label", "gn"),
        ("member_id", "r1i1p1f1"),
        ("source_id", "ACCESS-ESM1-5"),
        ("variable_id", "pr"),
    )
    assert executions[2].datasets[SourceDatasetType.PMPClimatology].selector == (
        ("source_id", "GPCP-Monthly-3-2"),
        ("variable_id", "pr"),
    )


@pytest.mark.parametrize(
    "variable_id,source_id,member_id",
    [
        ("pr", "ACCESS-ESM1-5", "r1i1p1f1"),
        ("ts", "MPI-ESM1-2-LR", "r2i2p1f1"),
    ],
)
def test_annual_cycle_diagnostic(
    variable_id,
    source_id,
    member_id,
    cmip6_data_catalog,
    obs4mips_data_catalog,
    definition_factory,
    pdo_example_dir,
    provider,
):
    diagnostic = AnnualCycle()
    diagnostic.provider = provider

    expected_input_filename = cmip6_data_catalog["path"].iloc[0]
    expected_reference_filename = obs4mips_data_catalog["path"].iloc[0]

    definition = definition_factory(
        diagnostic=diagnostic,
        cmip6=DatasetCollection(
            pd.Series(
                {
                    "instance_id": "HadISST",
                    "source_id": source_id,
                    "variable_id": variable_id,
                    "experiment_id": "historical",
                    "member_id": member_id,
                    "path": expected_input_filename,
                }
            )
            .to_frame()
            .T,
            "instance_id",
        ),
        pmp_climatology=DatasetCollection(
            pd.Series(
                {
                    "instance_id": "HadISST",
                    "source_id": "HadISST-1-1",
                    "variable_id": variable_id,
                    "path": expected_reference_filename,
                }
            )
            .to_frame()
            .T,
            "instance_id",
        ),
    )
    # The output directory must exist
    output_dir = definition.output_directory
    parameter_file = _get_resource(
        "climate_ref_pmp.params", "pmp_param_annualcycle_1-clims.py", use_resources=True
    )
    datecode = datetime.datetime.now().strftime("%Y%m%d")

    definition.output_directory.mkdir(parents=True)

    result = diagnostic.build_cmds(definition)

    assert len(result) == 2

    # Check the first command
    cmd = result[0]
    assert cmd == [
        "pcmdi_compute_climatologies.py",
        "-p",
        parameter_file,
        "--vars",
        variable_id,
        "--infile",
        str(expected_input_filename),
        "--outfile",
        f"{output_dir}/{variable_id}_{source_id}_historical_{member_id}_clims.nc",
    ]

    # Check the second command
    parameter_file = _get_resource(
        "climate_ref_pmp.params", "pmp_param_annualcycle_2-metrics.py", use_resources=True
    )
    cmd = result[1]
    assert cmd == [
        "mean_climate_driver.py",
        "-p",
        parameter_file,
        "--vars",
        variable_id,
        "--custom_observations",
        f"{output_dir}/obs_dict.json",
        "--test_data_path",
        str(output_dir),
        "--test_data_set",
        source_id,
        "--realization",
        member_id,
        "--filename_template",
        f"%(variable)_{source_id}_historical_{member_id}_clims.198101-200512.AC.v{datecode}.nc",
        "--metrics_output_path",
        str(output_dir),
        "--cmec",
    ]


def test_diagnostic_execute(mocker, provider):
    diagnostic = AnnualCycle()
    diagnostic.provider = provider

    # Mock the 'run' method of the 'provider' object
    mocked_provider_run = mocker.patch.object(
        provider,
        "run",
        autospec=True,
        spec_set=True,
    )

    diagnostic.build_cmds = mocker.MagicMock(return_value=[["mocked_command1"], ["mocked_command2"]])
    diagnostic.build_execution_result = mocker.MagicMock()

    # Create a mock ExecutionDefinition
    mock_definition = mocker.MagicMock(spec=ExecutionDefinition)
    diagnostic.execute(mock_definition)

    diagnostic.build_cmds.assert_called_once_with(mock_definition)
    assert mocked_provider_run.call_count == 2
    mocked_provider_run.assert_any_call(["mocked_command1"])
    mocked_provider_run.assert_any_call(["mocked_command2"])


def test_build_cmds(diagnostic_validation):
    diagnostic = AnnualCycle()
    diagnostic.provider = pmp_provider
    validator = diagnostic_validation(diagnostic)

    definition = validator.get_definition()
    definition.output_directory.mkdir(parents=True, exist_ok=True)

    cmds = diagnostic.build_cmds(definition)

    assert len(cmds) == 2

    assert (definition.output_directory / "obs_dict.json").exists()


def test_transform_results_removes_expected_keys():
    input_data = {
        "DIMENSIONS": {
            "json_structure": ["model", "reference", "rip", "region", "statistic", "season"],
            "model": {"ACCESS-ESM1-5": {}},
            "reference": {"default": {}},
            "rip": {"r1i1p1f1": {}},
            "region": {"global": {}, "NHEX": {}, "SHEX": {}, "TROPICS": {}},
            "statistic": {
                "bias_xy": {},
                "cor_xy": {},
                "mae_xy": {},
                "mean-obs_xy": {},
                "mean_xy": {},
                "rms_devzm": {},
                "rms_xy": {},
                "rms_xyt": {},
                "rms_y": {},
                "rmsc_xy": {},
                "std-obs_xy": {},
                "std-obs_xy_devzm": {},
                "std-obs_xyt": {},
                "std_xy": {},
                "std_xy_devzm": {},
                "std_xyt": {},
            },
            "season": {"ann": {}, "djf": {}, "mam": {}, "jja": {}, "son": {}, "CalendarMonths": {}},
        },
        "json_version": 3.0,
        "RESULTS": {
            "ACCESS-ESM1-5": {
                "default": {
                    "r1i1p1f1": {
                        "global": {
                            "bias_xy": {
                                "ann": 1.55238,
                                "djf": 1.66939,
                                "mam": 1.47425,
                                "jja": 1.47011,
                                "son": 1.59322,
                                "CalendarMonths": [
                                    "1.66806e+00",
                                    "1.62649e+00",
                                    "1.54286e+00",
                                    "1.40157e+00",
                                    "1.48475e+00",
                                    "1.55714e+00",
                                    "1.48418e+00",
                                    "1.36947e+00",
                                    "1.39818e+00",
                                    "1.57127e+00",
                                    "1.80950e+00",
                                    "1.71510e+00",
                                ],
                            },
                        },
                    },
                    "attributes": {"source": "ERA-5"},
                },
                "attributes": {"units": ""},
            }
        },
    }
    expected_output = {
        "DIMENSIONS": {
            "json_structure": ["region", "statistic", "season"],
            "region": {"NHEX": {}, "SHEX": {}, "TROPICS": {}, "global": {}},
            "season": {"ann": {}, "djf": {}, "jja": {}, "mam": {}, "son": {}},
            "statistic": {
                "bias_xy": {},
                "cor_xy": {},
                "mae_xy": {},
                "mean-obs_xy": {},
                "mean_xy": {},
                "rms_devzm": {},
                "rms_xy": {},
                "rms_xyt": {},
                "rms_y": {},
                "rmsc_xy": {},
                "std-obs_xy": {},
                "std-obs_xy_devzm": {},
                "std-obs_xyt": {},
                "std_xy": {},
                "std_xy_devzm": {},
                "std_xyt": {},
            },
        },
        "RESULTS": {
            "global": {
                "bias_xy": {"ann": 1.55238, "djf": 1.66939, "jja": 1.47011, "mam": 1.47425, "son": 1.59322}
            },
        },
        "json_version": 3.0,
    }
    transformed_data = _transform_results(input_data)
    assert transformed_data == expected_output


def test_transform_results_empty_results_and_dimensions():
    input_data = {
        "RESULTS": {},
        "DIMENSIONS": {"json_structure": []},
    }
    with pytest.raises(ValueError, match="'model' is not in list"):
        _transform_results(input_data)



#
#  OPTIONS ARE SET BY USER IN THIS FILE AS INDICATED BELOW BY:
#
#

# VARIABLES TO USE
# vars = ["rlut"]

# START AND END DATES FOR CLIMATOLOGY
start = "1981-01"
end = "2005-12"

# INPUT DATASET - CAN BE MODEL OR OBSERVATIONS
infile = (
    "obs4MIPs_PCMDI_monthly/NASA-LaRC/CERES-EBAF-4-1"
    "/mon/rlut/gn/v20210727"
    "/rlut_mon_CERES-EBAF-4-1_PCMDI_gn_200301-201812.nc"
)

# DIRECTORY WHERE TO PUT RESULTS
outfile = "climo/rlut_mon_CERES-EBAF-4-1_BE_gn.nc"



import pandas as pd
import pytest

from climate_ref.datasets.pmp_climatology import PMPClimatologyDatasetAdapter
from climate_ref_core.dataset_registry import dataset_registry_manager, fetch_all_files
from climate_ref_core.datasets import SourceDatasetType


@pytest.fixture(scope="session")
def pmp_data_catalog(sample_data_dir) -> pd.DataFrame:
    # Fetch the PMP data from the registry
    registry = dataset_registry_manager["pmp-climatology"]
    fetch_all_files(registry, "pmp-climatology", output_dir=None)

    # Ingest the data
    adapter = PMPClimatologyDatasetAdapter()
    return adapter.find_local_datasets(registry.abspath / "PMP_obs4MIPsClims")


@pytest.fixture(scope="session")
def data_catalog(cmip6_data_catalog, obs4mips_data_catalog, pmp_data_catalog):
    return {
        SourceDatasetType.CMIP6: cmip6_data_catalog,
        SourceDatasetType.obs4MIPs: obs4mips_data_catalog,
        SourceDatasetType.PMPClimatology: pmp_data_catalog,
    }



import pytest
from climate_ref_pmp import provider as pmp_provider
from climate_ref_pmp.diagnostics.enso import ENSO
from climate_ref_pmp.pmp_driver import _get_resource


def test_invalid_diagnostic():
    with pytest.raises(ValueError, match="Unknown metrics collection: invalid_metrics"):
        ENSO(metrics_collection="invalid_metrics")


def test_build_cmds(diagnostic_validation):
    diagnostic = ENSO("ENSO_tel")
    diagnostic.provider = pmp_provider
    validator = diagnostic_validation(diagnostic)

    definition = validator.get_definition()
    definition.output_directory.mkdir(parents=True, exist_ok=True)

    cmd = diagnostic.build_cmd(definition)

    assert cmd == [
        "python",
        _get_resource("climate_ref_pmp.drivers", "enso_driver.py", use_resources=True),
        "--metrics_collection",
        "ENSO_tel",
        "--experiment_id",
        "historical",
        "--input_json_path",
        str(definition.output_directory / "input_ENSO_tel_ACCESS-ESM1-5_historical_r1i1p1f1.json"),
        "--output_directory",
        str(definition.output_directory),
    ]



import pytest
from climate_ref_pmp.pmp_driver import build_glob_pattern, build_pmp_command, process_json_result

from climate_ref_core.pycmec.metric import CMECMetric
from climate_ref_core.pycmec.output import CMECOutput


def test_process_json_result(pdo_example_dir):
    json_file = (
        pdo_example_dir
        / "var_mode_PDO_EOF1_stat_cmip6_hist-GHG_mo_atm_ACCESS-ESM1-5_r1i1p1f1_2000-2005_cmec.json"
    )
    png_files = [pdo_example_dir / "pdo.png"]
    data_files = [pdo_example_dir / "pdo.nc"]

    cmec_output, cmec_metric = process_json_result(json_file, png_files, data_files)

    assert CMECMetric.model_validate(cmec_metric)
    assert CMECOutput.model_validate(cmec_output)
    assert len(cmec_metric.RESULTS)
    assert cmec_metric.DIMENSIONS.root["json_structure"] == [
        "model",
        "realization",
        "reference",
        "mode",
        "season",
        "method",
        "statistic",
    ]


def test_execute_missing_parameter():
    with pytest.raises(
        FileNotFoundError,
        match=r"Resource pmp_missing.py not found in climate_ref_pmp.params package.",
    ):
        build_pmp_command(
            driver_file="variability_mode/variability_modes_driver.py",
            parameter_file="pmp_missing.py",
            model_files=["model1.nc"],
            reference_name="HadISST-1-1",
            reference_paths=["reference.nc"],
            source_id="source_id",
            member_id="member_id",
            output_directory_path="output",
            experiment_id="historical",
        )


def test_build_glob_pattern_from_docstring_example():
    paths = [
        "/home/user/data/folder1/file1.nc",
        "/home/user/data/folder1/file2.nc",
        "/home/user/data/folder2/file3.nc",
    ]

    pattern = build_glob_pattern(paths)
    expected = "/home/user/data/**/file*.nc"
    assert pattern == expected


def test_build_glob_pattern_same_directory():
    paths = [
        "/home/user/data/folder1/sample_A.nc",
        "/home/user/data/folder1/sample_B.nc",
        "/home/user/data/folder1/sample_C.nc",
    ]

    pattern = build_glob_pattern(paths)
    expected = "/home/user/data/folder1/sample_*.nc"
    assert pattern == expected



import json
import os
from collections.abc import Collection, Iterable
from typing import Any

from loguru import logger

from climate_ref_core.constraints import AddSupplementaryDataset
from climate_ref_core.datasets import DatasetCollection, FacetFilter, SourceDatasetType
from climate_ref_core.diagnostics import (
    CommandLineDiagnostic,
    DataRequirement,
    ExecutionDefinition,
    ExecutionResult,
)
from climate_ref_pmp.pmp_driver import _get_resource, process_json_result


class ENSO(CommandLineDiagnostic):
    """
    Calculate the ENSO performance metrics for a dataset
    """

    facets = ("source_id", "member_id", "grid_label", "experiment_id", "metric", "reference_datasets")

    def __init__(self, metrics_collection: str, experiments: Collection[str] = ("historical",)) -> None:
        self.name = metrics_collection
        self.slug = metrics_collection.lower()
        self.metrics_collection = metrics_collection
        self.parameter_file = "pmp_param_enso.py"
        self.obs_sources: tuple[str, ...]
        self.model_variables: tuple[str, ...]

        if metrics_collection == "ENSO_perf":  # pragma: no cover
            self.model_variables = ("pr", "ts", "tauu")
            self.obs_sources = ("GPCP-Monthly-3-2", "TropFlux-1-0", "HadISST-1-1")
        elif metrics_collection == "ENSO_tel":
            self.model_variables = ("pr", "ts")
            self.obs_sources = ("GPCP-Monthly-3-2", "TropFlux-1-0", "HadISST-1-1")
        elif metrics_collection == "ENSO_proc":
            self.model_variables = ("ts", "tauu", "hfls", "hfss", "rlds", "rlus", "rsds", "rsus")
            self.obs_sources = (
                "GPCP-Monthly-3-2",
                "TropFlux-1-0",
                "HadISST-1-1",
                "CERES-EBAF-4-2",
            )
        else:
            raise ValueError(
                f"Unknown metrics collection: {metrics_collection}. "
                "Valid options are: ENSO_perf, ENSO_tel, ENSO_proc"
            )

        self.data_requirements = self._get_data_requirements(experiments)

    def _get_data_requirements(
        self,
        experiments: Collection[str] = ("historical",),
    ) -> tuple[DataRequirement, DataRequirement]:
        filters = [
            FacetFilter(
                facets={
                    "frequency": "mon",
                    "experiment_id": tuple(experiments),
                    "variable_id": self.model_variables,
                }
            )
        ]

        return (
            DataRequirement(
                source_type=SourceDatasetType.obs4MIPs,
                filters=(
                    FacetFilter(facets={"source_id": self.obs_sources, "variable_id": self.model_variables}),
                ),
                group_by=("activity_id",),
            ),
            DataRequirement(
                source_type=SourceDatasetType.CMIP6,
                filters=tuple(filters),
                group_by=("source_id", "experiment_id", "member_id", "grid_label"),
                constraints=(
                    AddSupplementaryDataset.from_defaults("areacella", SourceDatasetType.CMIP6),
                    AddSupplementaryDataset.from_defaults("sftlf", SourceDatasetType.CMIP6),
                ),
            ),
        )

    def build_cmd(self, definition: ExecutionDefinition) -> Iterable[str]:
        """
        Run the diagnostic on the given configuration.

        Parameters
        ----------
        definition : ExecutionDefinition
            The configuration to run the diagnostic on.

        Returns
        -------
        :
            The result of running the diagnostic.
        """
        mc_name = self.metrics_collection

        # ------------------------------------------------
        # Get the input datasets information for the model
        # ------------------------------------------------
        input_datasets = definition.datasets[SourceDatasetType.CMIP6]
        input_selectors = input_datasets.selector_dict()
        source_id = input_selectors["source_id"]
        member_id = input_selectors["member_id"]
        experiment_id = input_selectors["experiment_id"]
        variable_ids = set(input_datasets["variable_id"].unique()) - {"areacella", "sftlf"}
        mod_run = f"{source_id}_{member_id}"

        # We only need one entry for the model run
        dict_mod: dict[str, dict[str, Any]] = {mod_run: {}}

        def extract_variable(dc: DatasetCollection, variable: str) -> list[str]:
            return dc.datasets[input_datasets["variable_id"] == variable]["path"].to_list()  # type: ignore

        # TO DO: Get the path to the files per variable
        for variable in variable_ids:
            list_files = extract_variable(input_datasets, variable)
            list_areacella = extract_variable(input_datasets, "areacella")
            list_sftlf = extract_variable(input_datasets, "sftlf")

            if len(list_files) > 0:
                dict_mod[mod_run][variable] = {
                    "path + filename": list_files,
                    "varname": variable,
                    "path + filename_area": list_areacella,
                    "areaname": "areacella",
                    "path + filename_landmask": list_sftlf,
                    "landmaskname": "sftlf",
                }

        # -------------------------------------------------------
        # Get the input datasets information for the observations
        # -------------------------------------------------------
        reference_dataset = definition.datasets[SourceDatasetType.obs4MIPs]
        reference_dataset_names = reference_dataset["source_id"].unique()

        dict_obs: dict[str, dict[str, Any]] = {}

        # TO DO: Get the path to the files per variable and per source
        for obs_name in reference_dataset_names:
            dict_obs[obs_name] = {}
            for variable in variable_ids:
                # Get the list of files for the current variable and observation source
                list_files = reference_dataset.datasets[
                    (reference_dataset["variable_id"] == variable)
                    & (reference_dataset["source_id"] == obs_name)
                ]["path"].to_list()
                # If the list is not empty, add it to the dictionary
                if len(list_files) > 0:
                    dict_obs[obs_name][variable] = {
                        "path + filename": list_files,
                        "varname": variable,
                    }

        # Create input directory
        dict_datasets = {
            "model": dict_mod,
            "observations": dict_obs,
            "metricsCollection": mc_name,
            "experiment_id": experiment_id,
        }

        # Create JSON file for dictDatasets
        json_file = os.path.join(
            definition.output_directory, f"input_{mc_name}_{source_id}_{experiment_id}_{member_id}.json"
        )
        with open(json_file, "w") as f:
            json.dump(dict_datasets, f, indent=4)
        logger.debug(f"JSON file created: {json_file}")

        driver_file = _get_resource("climate_ref_pmp.drivers", "enso_driver.py", use_resources=True)
        return [
            "python",
            driver_file,
            "--metrics_collection",
            mc_name,
            "--experiment_id",
            experiment_id,
            "--input_json_path",
            json_file,
            "--output_directory",
            str(definition.output_directory),
        ]

    def build_execution_result(self, definition: ExecutionDefinition) -> ExecutionResult:
        """
        Build a diagnostic result from the output of the PMP driver

        Parameters
        ----------
        definition
            Definition of the diagnostic execution

        Returns
        -------
            Result of the diagnostic execution
        """
        input_datasets = definition.datasets[SourceDatasetType.CMIP6]
        source_id = input_datasets["source_id"].unique()[0]
        experiment_id = input_datasets["experiment_id"].unique()[0]
        member_id = input_datasets["member_id"].unique()[0]
        mc_name = self.metrics_collection
        pattern = f"{mc_name}_{source_id}_{experiment_id}_{member_id}"

        # Find the results files
        results_files = list(definition.output_directory.glob(f"{pattern}_cmec.json"))
        logger.debug(f"Results files: {results_files}")

        if len(results_files) != 1:  # pragma: no cover
            logger.warning(f"A single cmec output file not found: {results_files}")
            return ExecutionResult.build_from_failure(definition)

        # Find the other outputs
        png_files = [definition.as_relative_path(f) for f in definition.output_directory.glob("*.png")]
        data_files = [definition.as_relative_path(f) for f in definition.output_directory.glob("*.nc")]

        cmec_output, cmec_metric = process_json_result(results_files[0], png_files, data_files)

        input_selectors = definition.datasets[SourceDatasetType.CMIP6].selector_dict()
        cmec_metric_bundle = cmec_metric.remove_dimensions(
            [
                "model",
                "realization",
            ],
        ).prepend_dimensions(
            {
                "source_id": input_selectors["source_id"],
                "member_id": input_selectors["member_id"],
                "grid_label": input_selectors["grid_label"],
                "experiment_id": input_selectors["experiment_id"],
            }
        )

        return ExecutionResult.build_from_output_bundle(
            definition,
            cmec_output_bundle=cmec_output,
            cmec_metric_bundle=cmec_metric_bundle,
        )
